/*
 * Copyright 2016 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urmc.algo;

import java.io.PrintStream;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import at.jku.risc.stout.urmc.data.Hedge;
import at.jku.risc.stout.urmc.data.TermNode;
import at.jku.risc.stout.urmc.data.atom.ContextVar;
import at.jku.risc.stout.urmc.data.atom.Function;
import at.jku.risc.stout.urmc.data.atom.FunctionVar;
import at.jku.risc.stout.urmc.data.atom.HedgeVar;
import at.jku.risc.stout.urmc.data.atom.Hole;
import at.jku.risc.stout.urmc.data.atom.TermAtom;
import at.jku.risc.stout.urmc.data.atom.TermVar;

public class MatchAlgorithm {
	private final MatchSystem system;
	private final Set<Substitution> result = new HashSet<>();

	public MatchAlgorithm(MatchSystem system) {
		this.system = system;
		while (!system.getProblem().isEmpty()
				&& system.getProblem().getFirst().getLeft()
						.equals(system.getProblem().getFirst().getRight()))
			system.getProblem().remove();
	}

	public Set<Substitution> match(DebugLevel debugLevel, PrintStream debugOut,
			PrintStream debugErr, boolean justify) {
		MatchSystemCollection systems = new MatchSystemCollection();
		systems.add(system.copy());
		if (debugLevel == DebugLevel.PROGRESS)
			debugOut.println();

		while (!systems.isEmpty()) {
			MatchSystem sys = systems.remove();
			if (debugLevel == DebugLevel.PROGRESS)
				debugOut.println("Selected system: " + sys);
			while (!sys.getProblem().isEmpty()) {
				MatchEquation eq = sys.getProblem().remove();
				TermAtom leftA = eq.getLeft().getAtom();
				TermAtom rightA = eq.getRight().getAtom();
				Hedge leftH = eq.getLeft().getHedge();
				Hedge rightH = eq.getRight().getHedge();
				if (!leftA.equals(rightA)) {
					if (leftA instanceof TermVar && rightA != TermAtom.nullAtom) {
						// Individual Variable Elimination
						sys.apply((TermVar) leftA, eq.getRight());
						debug("Individual Variable Elimination", sys,
								debugLevel, debugOut, false);
					} else if (leftA instanceof FunctionVar
							&& (rightA instanceof Function || rightA instanceof FunctionVar)) {
						// Function Variable Elimination
						sys.getProblem().add(eq);
						sys.apply((FunctionVar) leftA, new TermNode(rightA,
								Hedge.nullHedge));
						debug("Function Variable Elimination", sys, debugLevel,
								debugOut, false);
					} else if (leftA instanceof ContextVar) {
						// Deepening
						if (rightA != TermAtom.nullAtom)
							for (int i = 0, n = rightH.size(); i < n; i++) {
								MatchSystem sysNew = sys.copy();
								TermNode leftNew = eq.getLeft().copy();
								TermNode rightNew = eq.getRight().copy();
								TermNode idxNode = rightNew.getHedge().get(i);
								rightNew.getHedge()
										.getSequence()
										.set(i,
												new TermNode(leftA,
														Hole.INSTANCE.hedge));
								sysNew.apply((ContextVar) leftA, rightNew);
								leftNew.setHedge(leftNew.getHedge().substitute(
										(ContextVar) leftA, rightNew));
								sysNew.getProblem().add(
										new MatchEquation(leftNew, idxNode));
								systems.add(sysNew);
								debug("Deepening", sysNew, debugLevel,
										debugOut, true);
							}
						// Context Variable Deletion
						sys.getProblem().add(eq);
						sys.apply((ContextVar) leftA, Hole.INSTANCE.node);
						debug("Context Variable Deletion", sys, debugLevel,
								debugOut, false);
					} else {
						// Symbol Clash
						sys.getProblem().add(eq);
						error("Symbol Clash", sys, debugLevel, debugErr);
						break;
					}
				} else if (leftH.equals(rightH)) {
					// Trivial
					debug("Trivial", sys, debugLevel, debugOut, false);
				} else if (leftH.size() == 0) {
					// Arity Disagreement because it follows rightH.size() > 0
					sys.getProblem().add(eq);
					error("Arity Disagreement", sys, debugLevel, debugErr);
					break;
				} else if (leftH.get(0).getAtom() instanceof HedgeVar) {
					HedgeVar leftVar = (HedgeVar) leftH.get(0).getAtom();
					// Widening
					if (rightH.size() > 0) {
						MatchSystem sysNew = sys.copy();
						MatchEquation eqNew = eq.copy();
						List<TermNode> leftSeq = eqNew.getLeft().getHedge()
								.getSequence();
						TermNode x = leftSeq.remove(0);
						Hedge widen = new Hedge();
						widen.add(eqNew.getRight().getHedge().getSequence()
								.remove(0)); // t
						widen.add(x);
						sysNew.getProblem().add(eqNew);
						sysNew.apply(leftVar, new TermNode(null, widen));
						leftSeq.add(0, x);
						systems.add(sysNew);
						debug("Widening", sysNew, debugLevel, debugOut, true);
					}
					// Sequence Variable Deletion
					sys.getProblem().add(eq);
					sys.apply(leftVar, TermNode.EMPTY);
					debug("Sequence Variable Deletion", sys, debugLevel,
							debugOut, false);
				} else {
					int idx = idxHedgeVar(leftH, rightH);
					if (idx != -1) {
						// Partial Decomposition 1
						for (int i = 0; i < idx; i++)
							sys.getProblem().add(
									new MatchEquation(leftH.get(i), rightH
											.get(i)));
						eq.setLeft(new TermNode(leftA, leftH.subHedge(idx,
								leftH.size())));
						eq.setRight(new TermNode(rightA, rightH.subHedge(idx,
								rightH.size())));
						sys.getProblem().add(eq);
						debug("Partial Decomposition", sys, debugLevel,
								debugOut, false);
					} else if (leftH.size() == rightH.size()) {
						// Total Decomposition
						for (int i = leftH.size() - 1; i >= 0; i--)
							sys.getProblem().add(
									new MatchEquation(leftH.get(i), rightH
											.get(i)));

						if (leftA != TermAtom.nullAtom || leftH.size() != 1)
							debug("Total Decomposition", sys, debugLevel,
									debugOut, false); // Non-trivial
					} else {
						// Arity Disagreement
						sys.getProblem().add(eq);
						error("Arity Disagreement", sys, debugLevel, debugErr);
						break;
					}
				}
			}
			// Empty problem set = solution found
			if (sys.getProblem().isEmpty()) {
				if (justify) {
					MatchProblem problemOrig = system.getProblem().copy();
					while (!problemOrig.isEmpty()) {
						MatchEquation eqOrig = problemOrig.remove();
						eqOrig.apply(sys.getSigma());
						if (!eqOrig.getLeft().equals(eqOrig.getRight())) {
							debugErr.println("ERROR: Justification for the following matcher failed!!!");
							debugErr.println("ERROR: Matcher: "
									+ sys.getSigma());
							return null;
						}
					}
					if (debugLevel == DebugLevel.PROGRESS) {
						debugOut.println("  Justification OK for matcher: "
								+ sys.getSigma());
						debugOut.println();
					}
				}
				result.add(sys.getSigma());
			}
		}

		if (debugLevel != DebugLevel.SILENT) {
			if (result.isEmpty())
				debugOut.println("No matcher found.");
			else
				debugOut.println("Matchers found: " + result);
		}

		return result;
	}

	private void debug(String msg, MatchSystem sys, DebugLevel debugLevel,
			PrintStream out, boolean newBranch) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("  ==>  " + msg + " ==>");
			if (newBranch)
				out.println("  New branch generated!");
		}
		debug(debugLevel, out, sys);
	}

	private void error(String msg, MatchSystem sys, DebugLevel debugLevel,
			PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("  ==>  " + msg);
			out.println();
		}
	}

	private void debug(DebugLevel debugLevel, PrintStream out, MatchSystem sys) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("  System: " + sys);
			out.println();
		}
	}

	private int idxHedgeVar(Hedge leftH, Hedge rightH) {
		for (int n = Math.min(leftH.size(), rightH.size()), i = 0; i < n; i++)
			if (leftH.get(i).getAtom() instanceof HedgeVar)
				return i;
		return -1;
	}

	public MatchSystem getInitialSystem() {
		return system;
	}

	public Set<Substitution> getResult() {
		return result;
	}
}
